@@ -2,7 +2,7 @@
 ## Name:        Smart.pm
 ## Purpose:     XML::Smart
 ## Author:      Graciliano M. P.
-## Modified by:
+## Modified by: Harish Madabushi
 ## Created:     10/05/2003
 ## RCS-ID:      
 ## Copyright:   (c) 2003 Graciliano M. P.
@@ -10,29 +10,47 @@
 ##              modify it under the same terms as Perl itself
 #############################################################################
 
-package XML::Smart ;
-use 5.006 ;
 
-no warnings ;
+package XML::Smart                                             ;
 
-use Object::MultiType ;
-use vars qw(@ISA) ;
-@ISA = qw(Object::MultiType) ;
+use 5.006                                                      ; 
+
+use strict                                                     ;
+use warnings                                                   ;
+
+use Object::MultiType                                          ;
+
+use XML::Smart::Shared qw( _unset_sig_warn _reset_sig_warn )   ;
+
+use vars qw(@ISA)                                              ;
+@ISA = qw(Object::MultiType)                                   ;
+
+use XML::Smart::Tie                                            ;
+use XML::Smart::Tree                                           ;
 
-use XML::Smart::Tie ;
-use XML::Smart::Tree ;
 
 our ($VERSION) ;
-$VERSION = '1.6' ;
+$VERSION = '1.71' ;
+
 
 ###############
 # AUTOLOADERS #
 ###############
 
 sub data {
-  require XML::Smart::Data ;
-  *data = \&XML::Smart::Data::data ;
-  &XML::Smart::Data::data(@_) ;
+    require XML::Smart::Data ;
+    _unset_sig_warn() ;
+    *data = \&XML::Smart::Data::data ;
+    _reset_sig_warn() ;
+    &XML::Smart::Data::data(@_) ;
+}
+
+sub apply_dtd {
+    require XML::Smart::DTD ;
+    _unset_sig_warn() ;
+    *apply_dtd = \&XML::Smart::DTD::apply_dtd ;
+    _reset_sig_warn() ;
+    &XML::Smart::DTD::apply_dtd(@_) ;
 }
 
 sub xpath { _load_xpath() ; &XML::Smart::XPath::xpath(@_) ;}
@@ -41,12 +59,15 @@ sub xpath_pointer { _load_xpath() ; &XML::Smart::XPath::xpath_pointer(@_) ;}
 sub XPath_pointer { _load_xpath() ; &XML::Smart::XPath::XPath_pointer(@_) ;}
 
 sub _load_xpath {
-  require XML::Smart::XPath ;
-  *xpath = \&XML::Smart::XPath::xpath ;
-  *XPath = \&XML::Smart::XPath::XPath ;
-  *xpath_pointer = \&XML::Smart::XPath::xpath_pointer ;
-  *XPath_pointer = \&XML::Smart::XPath::XPath_pointer ;
-  *_load_xpath = sub {} ;
+    require XML::Smart::XPath ;
+    _unset_sig_warn() ;
+    *xpath = \&XML::Smart::XPath::xpath ;
+    *XPath = \&XML::Smart::XPath::XPath ;
+    *xpath_pointer = \&XML::Smart::XPath::xpath_pointer ;
+    *XPath_pointer = \&XML::Smart::XPath::XPath_pointer ;
+    *_load_xpath = sub {} ;
+    _reset_sig_warn() ;
+
 }
 
 #################
@@ -54,7 +75,7 @@ sub _load_xpath {
 #################
 
 sub NO_XML_PARSER {
-  $XML::Smart::Tree::NO_XML_PARSER = !@_ ? 1 : ( $_[0] ? 1 : undef ) ;
+    $XML::Smart::Tree::NO_XML_PARSER = !@_ ? 1 : ( $_[0] ? 1 : undef ) ;
 }
 
 #######
@@ -62,27 +83,28 @@ sub NO_XML_PARSER {
 #######
 
 sub new {
-  my $class = shift ;
-  my $file = shift ;
-  my $parser = ($_[0] !~ /^(?:uper|low|arg|on)\w+$/i) ? shift(@_) : '' ;
-  
-  my $this = Object::MultiType->new(
-  boolsub   => \&boolean ,
-  scalarsub => \&content ,
-  tiearray  => 'XML::Smart::Tie::Array' ,
-  tiehash   => 'XML::Smart::Tie::Hash' ,
-  tieonuse  => 1 ,
-  code      => \&find_arg , 
-  ) ;
-  
-  my $parser = &XML::Smart::Tree::load($parser) ;
-  
-  if ($file eq '') { $$this->{tree} = {} ;}
-  else { $$this->{tree} = &XML::Smart::Tree::parse($file,$parser,@_) ;}
 
-  $$this->{point} = $$this->{tree} ;
-  
-  bless($this,$class) ;
+    my $class = shift ;
+    my $file  = shift ;
+    my $parser = ($_[0] and $_[0] !~ /^(?:uper|low|arg|on|no|use)\w+$/i) ? shift(@_) : '' ;
+    
+    my $this = Object::MultiType->new(
+	boolsub   => \&boolean ,
+	scalarsub => \&content ,
+	tiearray  => 'XML::Smart::Tie::Array' ,
+	tiehash   => 'XML::Smart::Tie::Hash' ,
+	tieonuse  => 1 ,
+	code      => \&find_arg , 
+	) ;
+    
+    $parser = &XML::Smart::Tree::load($parser) ;
+    
+    if ( !($file) or $file eq '') { $$this->{tree} = {} ;}
+    else { $$this->{tree} = &XML::Smart::Tree::parse($file,$parser,@_) ;}
+    
+    $$this->{point} = $$this->{tree} ;
+    
+    bless($this,$class) ;
 }
 
 #########
@@ -90,83 +112,84 @@ sub new {
 #########
 
 sub clone {
-  my $saver = shift ;
-
-  my ($pointer , $back , $array , $key , $i , $null_clone) ;
-
-  if ($#_ == 0 && !ref $_[0]) {
-    my $nullkey = shift ;
-    $pointer = {} ;
-    $back = {} ;
-    $null_clone = 1 ;
+    my $saver = shift ;
+    
+    my ($pointer , $back , $array , $key , $i , $null_clone) ;
+    
+    if ($#_ == 0 && !ref $_[0]) {
+	my $nullkey = shift ;
+	$pointer = {} ;
+	$back = {} ;
+	$null_clone = 1 ;
+	
+	($i) = ( $nullkey =~ /(?:^|\/)\/\[(\d+)\]$/s );
+	($key) = ( $nullkey =~ /(.*?)(?:\/\/\[\d+\])?$/s );
+	if ($key =~ /^\/\[\d+\]$/) { $key = undef ;}
+    }
+    
+    else {
+	$pointer = shift ;
+	$back = shift ;
+	$array = shift ;
+	$key = shift ;
+	$i = shift ;
+    }
     
-    ($i) = ( $nullkey =~ /(?:^|\/)\/\[(\d+)\]$/s );
-    ($key) = ( $nullkey =~ /(.*?)(?:\/\/\[\d+\])?$/s );
-    if ($key =~ /^\/\[\d+\]$/) { $key = undef ;}
-  }
-
-  else {
-    $pointer = shift ;
-    $back = shift ;
-    $array = shift ;
-    $key = shift ;
-    $i = shift ;
-  }
-
   my $clone = Object::MultiType->new(
-  boolsub   => \&boolean ,
+      boolsub   => \&boolean ,
   scalarsub => \&content ,
-  tiearray  => 'XML::Smart::Tie::Array' ,
-  tiehash   => 'XML::Smart::Tie::Hash' ,
-  tieonuse  => 1 ,
-  code      => \&find_arg ,
-  ) ;
-  bless($clone,__PACKAGE__) ;  
-  
-  if ( !$saver->is_saver ) { $saver = $$saver ;}
-  
-  if (!$back) {
-    if (!$pointer) { $back = $saver->{back} ;}
-    else { $back = $saver->{point} ;}
-  }
-  
-  if (!$array && !$pointer) { $array = $saver->{array} ;}
-
-  my @keyprev ;
-
-  if (defined $key) { @keyprev = $key ;}
-  elsif (defined $i) { @keyprev = "[$i]" ;}
-
-  if (!defined $key) { $key = $saver->{key} ;}
-  if (!defined $i) { $i = $saver->{i} ;}
+      tiearray  => 'XML::Smart::Tie::Array' ,
+      tiehash   => 'XML::Smart::Tie::Hash' ,
+      tieonuse  => 1 ,
+      code      => \&find_arg ,
+      ) ;
+    bless($clone,__PACKAGE__) ;  
   
+    if ( !$saver->is_saver ) { $saver = $$saver ;}
+    
+    if (!$back) {
+	if (!$pointer) { $back = $saver->{back} ;}
+	else { $back = $saver->{point} ;}
+    }
+    
+    if (!$array && !$pointer) { $array = $saver->{array} ;}
+    
+    my @keyprev ;
+    
+    if (defined $key) { @keyprev = $key ;}
+    elsif (defined $i) { @keyprev = "[$i]" ;}
+    
+    if (!defined $key) { $key = $saver->{key} ;}
+    if (!defined $i) { $i = $saver->{i} ;}
+    
   if (!$pointer) { $pointer = $saver->{point} ;}
-  
-  #my @call = caller ;
-  #print "CLONE>> $key , $i >> @{$saver->{keyprev}} >> @_\n" ;
-
-  $$clone->{tree} = $saver->{tree} ;
-  $$clone->{point} = $pointer ;
-  $$clone->{back} = $back ;
-  $$clone->{array} = $array ;
-  $$clone->{key} = $key ;
-  $$clone->{i} = $i ;
-  
-  if ( @keyprev ) {
-    $$clone->{keyprev} = [@{$saver->{keyprev}}] ;
-    push(@{$$clone->{keyprev}} , @keyprev) ;
-  }
-  
-  if (defined $_[0]) { $$clone->{content} = \$_[0] ;}
-
-  if ( $null_clone || defined $saver->{null} ) {
-    $$clone->{null} = 1 ;
-    ## $$clone->{self} = $clone ;
-  }
-  
-  $$clone->{XPATH} = $saver->{XPATH} if $saver->{XPATH} ;
-  
+    
+    #my @call = caller ;
+    #print "CLONE>> $key , $i >> @{$saver->{keyprev}} >> @_\n" ;
+    
+    $$clone->{tree} = $saver->{tree} ;
+    $$clone->{point} = $pointer ;
+    $$clone->{back} = $back ;
+    $$clone->{array} = $array ;
+    $$clone->{key} = $key ;
+    $$clone->{i} = $i ;
+    
+    if ( @keyprev ) {
+	$$clone->{keyprev} = ( $saver->{keyprev} ) ? [@{$saver->{keyprev}}]  : [] ;
+	push(@{$$clone->{keyprev}} , @keyprev) ;
+    }
+    
+    if (defined $_[0]) { $$clone->{content} = \$_[0] ;}
+    
+    if ( $null_clone || $saver->{null} ) {
+	$$clone->{null} = 1 ;
+	## $$clone->{self} = $clone ;
+    }
+    
+    $$clone->{XPATH} = $saver->{XPATH} if $saver->{XPATH} ;
+    
   return( $clone ) ;
+
 }
 
 ###########
@@ -324,7 +347,7 @@ sub key {
 
 sub i {
   my $this = shift ;
-  my $i = $$this->{i} || undef ;
+  my $i = $$this->{i} ;
   return $i ;
 }
 
@@ -461,11 +484,13 @@ sub tree_pointer { &pointer ;}
 
 sub dump_tree {
   require Data::Dumper ;
+  local $Data::Dumper::Sortkeys = 1 ;
   return Data::Dumper::Dumper( &tree ) ;
 }
 
 sub dump_tree_ok {
   require Data::Dumper ;
+  local $Data::Dumper::Sortkeys = 1 ;
   return Data::Dumper::Dumper( &tree_ok ) ;
 }
 
@@ -476,11 +501,13 @@ sub dump_tree_ok {
 
 sub dump_pointer {
   require Data::Dumper ;
+  local $Data::Dumper::Sortkeys = 1 ;
   return Data::Dumper::Dumper( &pointer ) ;
 }
 
 sub dump_pointer_ok {
   require Data::Dumper ;
+  local $Data::Dumper::Sortkeys = 1 ;
   return Data::Dumper::Dumper( &pointer_ok ) ;
 }
 
@@ -508,7 +535,7 @@ sub cut_root {
 
   if ($#nodes > 0) { return $this ;}
   
-  my $root = @nodes[0] ;
+  my $root = $nodes[0] ;
   return( $this->{$root} ) ;
 }
 
@@ -518,6 +545,8 @@ sub cut_root {
 
 sub is_node {
   my $this = shift ;
+  return if $this->null ;
+  
   my $key = $this->key ;
   
   my $back = $this->back ;
@@ -532,6 +561,8 @@ sub is_node {
 
 sub args {
   my $this = shift ;
+  return () if $this->null ;
+  
   my @args ;
 
   my $nodes = $this->back->{'/nodes'} ;
@@ -554,6 +585,9 @@ sub args {
 
 sub args_values {
   my $this = shift ;
+  
+  return () if $this->null ;
+  
   my @args = $this->args ;
 
   my @values ;
@@ -572,6 +606,8 @@ sub args_values {
 sub nodes {
   my $this = shift ;
 
+  return () if $this->null ;
+    
   my $nodes = $this->{'/nodes'}->pointer ;
   my $pointer = $$this->{point} ;
   
@@ -600,6 +636,8 @@ sub nodes {
 
 sub nodes_keys {
   my $this = shift ;
+  
+  return () if $this->null ;
 
   my $nodes = $this->{'/nodes'}->pointer ;
   my $pointer = $$this->{point} ;
@@ -632,20 +670,23 @@ sub set_node {
   my $nodes = $back->{'/nodes'}->pointer ;
   
   if ( $bool ) {
-    if ( $$nodes{$key} =~ /^(\w+,\d+),(\d*)/ ) { $$nodes{$key} = "$1,1" ;}
-    else { $$nodes{$key} = 1 ;}
 
-    if ( !$this->{CONTENT} ) {
-      my $content = $this->content ;
-      $this->{CONTENT} = $content if $content ne '' ;
-    }
-  }
-  else {
-    delete $$nodes{$key} ;
-    my @keys = keys %$this ;
-    if ( $#keys == 0 && @keys[0] eq 'CONTENT') {
-      my $content = $this->{CONTENT}('.') ;
-      $this->back->pointer->{$key} = $content ;
+      if ( $$nodes{$key} && $$nodes{$key} =~ /^(\w+,\d+),(\d*)/ ) { 
+	  $$nodes{$key} = "$1,1" ;
+      }else { 
+	  $$nodes{$key} = 1 ;
+      }
+      
+      if ( !$this->{CONTENT} ) {
+	  my $content = $this->content ;
+	  $this->{CONTENT} = $content if $content ne '' ;
+      }
+  } else {
+      delete $$nodes{$key} ;
+      my @keys = keys %$this ;
+      if ( $#keys == 0 && $keys[0] eq 'CONTENT') {
+	  my $content = !$this->{CONTENT}->null ? $this->{CONTENT}('.') : $this->content ;
+	  $this->back->pointer->{$key} = $content ;
     }
   }
   
@@ -679,45 +720,48 @@ sub order {
 #############
 
 sub set_node_type {
-  my $this = shift ;
-  my ( $type , $bool ) = @_ ;
-  if ( $#_ < 1 ) { $bool = 1 ;}
-  
-  my $key = $this->key ;
-  
-  my $back = $this->back ;
-  
-  $back->{'/nodes'} = {} if $back->{'/nodes'}->null ;
-  my $nodes = $back->{'/nodes'}->pointer ;
-  
-  if ( $bool ) {
-    if ( $$nodes{$key} =~ /^\w+,\d+,(\d*)/ ) {
-      my $val = $1 ;
-      $$nodes{$key} = "$type,1,$val" ;
-    }
-    else { $$nodes{$key} = "$type,1,$$nodes{$key}" ;}
 
-    if ( !$this->{CONTENT} ) {
-      my $content = $this->content ;
-      $this->{CONTENT} = $content if $content ne '' ;
+    my $this = shift ;
+    my ( $type , $bool ) = @_ ;
+    if ( $#_ < 1 ) { $bool = 1 ;}
+    
+    my $key = $this->key ;
+    
+    my $back = $this->back ;
+    
+    $back->{'/nodes'} = {} if $back->{'/nodes'}->null ;
+    my $nodes = $back->{'/nodes'}->pointer ;
+    
+    if ( $bool ) {
+	if ( $$nodes{$key} && $$nodes{$key} =~ /^\w+,\d+,(\d*)/ ) {
+	    my $val = $1                   ;
+	    $$nodes{$key} = "$type,1,$val" ;
+	} else { 
+	    my $existing_node_data = ( $$nodes{$key} ) ? $$nodes{$key} : "" ;
+	    $$nodes{$key} = "$type,1," . $existing_node_data ;
+	}
+	
+	if ( !$this->{CONTENT} ) {
+	    my $content = $this->content ;
+	    $this->{CONTENT} = $content if $content ne '' ;
+	}
     }
-  }
-  else {
-    if ( !$$nodes{$key} ) {
-      my $tp = _data_type( $back->{$key} ) ;
-      if ( $tp > 2 ) { $$nodes{$key} = "$type,0," ;}
+    else {
+	if ( !$$nodes{$key} ) {
+	    my $tp = _data_type( $back->{$key} ) ;
+	    if ( $tp > 2 ) { $$nodes{$key} = "$type,0," ;}
+	}
+	elsif ( $$nodes{$key} eq '1' ) { $$nodes{$key} = "$type,0,1" ;}
+	elsif ( $$nodes{$key} =~ /^\w+,\d+,1/ ) { $$nodes{$key} = "$type,0,1" ;}
+	elsif ( $$nodes{$key} =~ /^\w+,\d+,0?$/ ) {
+	    delete $$nodes{$key} ;
+	    my @keys = keys %$this ;
+	    if ( $#keys == 0 && $keys[0] eq 'CONTENT') {
+		my $content = $this->{CONTENT}('.') ;
+		$this->back->pointer->{$key} = $content ;
+	    }
+	}
     }
-    elsif ( $$nodes{$key} eq '1' ) { $$nodes{$key} = "$type,0,1" ;}
-    elsif ( $$nodes{$key} =~ /^\w+,\d+,1/ ) { $$nodes{$key} = "$type,0,1" ;}
-    elsif ( $$nodes{$key} =~ /^\w+,\d+,0?$/ ) {
-      delete $$nodes{$key} ;
-      my @keys = keys %$this ;
-      if ( $#keys == 0 && @keys[0] eq 'CONTENT') {
-        my $content = $this->{CONTENT}('.') ;
-        $this->back->pointer->{$key} = $content ;
-      }
-    }
-  }
 }
 
 #############
@@ -756,7 +800,7 @@ sub set_auto_node {
   elsif ( $$nodes{$key} =~ /^\w+,\d+,0?$/ ) {
     delete $$nodes{$key} ;
     my @keys = keys %$this ;
-    if ( $#keys == 0 && @keys[0] eq 'CONTENT') {
+    if ( $#keys == 0 && $keys[0] eq 'CONTENT') {
       my $content = $this->{CONTENT}('.') ;
       $this->back->pointer->{$key} = $content ;
     }
@@ -778,7 +822,7 @@ sub set_auto {
   
   delete $$nodes{$key} ;
   my @keys = keys %$this ;
-  if ( $#keys == 0 && @keys[0] eq 'CONTENT') {
+  if ( $#keys == 0 && $keys[0] eq 'CONTENT') {
     my $content = $this->{CONTENT}('.') ;
     $this->back->pointer->{$key} = $content ;
   }
@@ -794,10 +838,10 @@ sub set_auto {
 ## 1 value
 
 sub _data_type {
-  return 4 if ($_[0] =~ /[^\w\d\s!"#\$\%&'\(\)\*\+,\-\.\/:;<=>\?\@\[\\\]\^\`\{\|}~]/s) ;
-  return 3 if ($_[0] =~ /<.*?>/s) ;
-  return 2 if ($_[0] =~ /[\r\n\t]/s) ;
-  return 1 ;
+    return 4 if( $_[0] && $_[0] =~ /[^\w\d\s!"#\$\%&'\(\)\*\+,\-\.\/:;<=>\?\@\[\\\]\^\`\{\|}~]/s) ;
+    return 3 if( $_[0] && $_[0] =~ /<.*?>/s    ) ;
+    return 2 if( $_[0] && $_[0] =~ /[\r\n\t]/s ) ;
+    return 1 ;
 }
 
 #######
@@ -927,6 +971,7 @@ sub find_arg {
       $data = ref($hash) eq 'HASH' ? $$hash{$name} : $hash ;
       $data = $$data{CONTENT} if ref($data) eq 'HASH' ;
       
+      _unset_sig_warn() ;
       if    ($type eq 'eq'  && $data eq $value)     { push(@hash,$hash_i) ; push(@i,$i) ; last ;}
       elsif ($type eq 'ne'  && $data ne $value)     { push(@hash,$hash_i) ; push(@i,$i) ; last ;}
       elsif ($type eq '=='  && $data == $value)     { push(@hash,$hash_i) ; push(@i,$i) ; last ;}
@@ -939,6 +984,7 @@ sub find_arg {
       elsif ($type eq '=~i' && $data =~ /$value/is) { push(@hash,$hash_i) ; push(@i,$i) ; last ;}
       elsif ($type eq '!~'  && $data !~ /$value/s)  { push(@hash,$hash_i) ; push(@i,$i) ; last ;}
       elsif ($type eq '!~i' && $data !~ /$value/is) { push(@hash,$hash_i) ; push(@i,$i) ; last ;}
+      _reset_sig_warn() ;
     }
 
     if ($notwant && @hash) { last ;}
@@ -980,6 +1026,12 @@ sub content {
   my $this = shift ;
   my $set_i = $#_ > 0 ? shift : undef ;
   
+  if ( $this->null ) {
+    &XML::Smart::Tie::_generate_nulltree( $$this ) ;
+  }
+  
+  ##use Data::Dumper; print Dumper($$this) ;
+  
   if ( defined $$this->{content} ) {
     if (@_) { ${$$this->{content}} = $_[0] ;}
     return ${$$this->{content}} ;
@@ -988,15 +1040,27 @@ sub content {
   my $key = 'CONTENT' ;
   my $i = $$this->{i} ;
   
+  if (ref($$this->{point}) eq 'ARRAY') {
+    return $this->[0]->content($set_i,@_) ;
+  }
+  
   if ( ref($$this->{point}) ne 'HASH' ) { return '' ;}
   
+  if ( !exists $$this->{point}{$key} ) {
+    if ( @_ ) { return $$this->{point}{$key} = $_[0] ;}
+    return '' ;
+  }
+  
   if (ref($$this->{point}{$key}) eq 'ARRAY') {
     if ($i eq '') { $i = 0 ;}
     if (@_) { $$this->{point}{$key}[$i] = $_[0] ;}
     return $$this->{point}{$key}[$i] ;
   }
   elsif (exists $$this->{point}{$key}) {
-    if (@_ && ( my $tie = tied($$this->{point}{$key}) ) ) { $tie->STORE($set_i , $_[0]) ;}
+    if ( @_ ) {
+      if ( my $tie = tied($$this->{point}{$key}) ) { $tie->STORE($set_i , $_[0]) ;}
+      else { $$this->{point}{$key} = $_[0] ;}
+    }
     if ( wantarray && ( my $tie = tied($$this->{point}{$key}) ) ) { return $tie->FETCH(1) ;}
     return $$this->{point}{$key} ;
   }
@@ -1056,7 +1120,27 @@ sub data_pointer {
 
 sub DESTROY {
   my $this = shift ;
-  $$this->clean ;
+  $$this->clean if( $this && $$this ) ; # In case object was messed with ( bug 62091 ) 
+}
+
+###################
+# STORABLE_FREEZE #
+###################
+
+sub STORABLE_freeze {
+  my $this = shift ;
+  return($this , [$$this->{tree} , $$this->{pointer}])  ;
+}
+
+#################
+# STORABLE_THAW #
+#################
+
+sub STORABLE_thaw {
+  my $this = shift ;
+  $$this->{tree} = $_[1]->[0] ;
+  $$this->{pointer} = $_[1]->[1] ;
+  return ;
 }
 
 #######
@@ -1073,19 +1157,20 @@ XML::Smart - A smart, easy and powerful way to access/create XML files/data.
 
 =head1 DESCRIPTION
 
-This module has an easy way to access/create XML data. It's based on the HASH
-tree that is made of the XML data, and enable a dynamic access to it with the
-Perl syntax for Hash and Array, without needing to care if you have a Hash or an
-Array in the tree. In other words, B<each point in the tree work as a Hash and
+This module provides an easy way to access/create XML data. It's based on a HASH
+tree created from the XML data, and enables dynamic access to it through the 
+standard Perl syntax for Hash and Array, without necessarily caring about which 
+you are working with. In other words, B<each point in the tree works as a Hash and
 an Array at the same time>!
 
-You also have extra resources, like a search for nodes by attribute, selection
-of an attribute value in each multiple node,  change the returned format, etc...
+This module additionally provides special resources such as: search for nodes by 
+attribute, select an attribute value in each multiple node,  change the returned 
+format, and so on.
 
-The module alson handle automatically binary data (encoding/decoding to/from base64),
+The module also automatically handles binary data (encoding/decoding to/from base64),
 CDATA (like contents with <tags>) and Unicode. It can be used to create XML files,
-load XML from the Web (just pasting an URL as a file path) and it has an easy
-way to send XML data through socket, just adding the length of the data in
+load XML from the Web ( just by using an URL as the file path ) and has an easy
+way to send XML data through sockets - just adding the length of the data in
 the <?xml?> header.
 
 You can use I<XML::Smart> with L<XML::Parser>, or with the 2 standart parsers of
@@ -1101,6 +1186,20 @@ XML::Smart:
 
 I<XML::Smart::HTMLParser> can be used to load/parse wild/bad XML data, or HTML tags.
 
+=head1 Tutorial and F.A.Q.
+
+You can find some extra documents about I<XML::Smart> at:
+
+=over 2
+
+=item L<XML::Smart::Tutorial> - Tutorial and examples for XML::Smart.
+
+=item L<XML::Smart::FAQ>      - Frequently Asked Questions about XML::Smart.
+
+=back
+
+=cut
+
 =head1 USAGE
 
   ## Create the object and load the file:
@@ -1192,7 +1291,7 @@ The first argument can be:
   - File Handle (GLOB).
   - URL (Need LWP::UserAgent).
 
-If not paste, a null XML tree is started, where you should create your own
+If not passed, a null XML tree is started, where you should create your own
 XML data, than build/save/send it.
 
 =item PARSER B<(optional)>
@@ -1217,7 +1316,7 @@ Aliases for the options:
 I<Default:>
 
 If not set it will look for XML::Parser and load it.
-If XML::Parser can't be loaded it will use XML::Smart::Parser, that actually is a
+If XML::Parser can't be loaded it will use XML::Smart::Parser, which is actually a
 clone of XML::Parser::Lite with some fixes.
 
 =item OPTIONS
@@ -1265,6 +1364,10 @@ Here's the tree of the example above:
                        },
             },
 
+=item use_spaces
+
+Accept contents that have only spaces.
+
 =item on_start (CODE) I<*optional>
 
 Code/sub to call on start a tag.
@@ -1318,6 +1421,45 @@ B<Examples of use:>
              }
   ) ;
 
+=head2  apply_dtd (DTD , OPTIONS)
+
+Apply the I<DTD> to the XML tree.
+
+I<DTD> can be a source, file, GLOB or URL.
+
+This method is usefull if you need to have the XML generated by I<data()>
+formated in a specific DTD, so, elements will be nodes automatically,
+attributes will be checked, required elements and attributes will be created,
+the element order will be set, etc...
+
+B<OPTIONS:>
+
+=over 10
+
+=item no_delete BOOL
+
+If TRUE tells that not defined elements and attributes in the DTD won't be deleted
+from the XML tree.
+
+=back
+
+B<Example of use:>
+
+  $xml->apply_dtd(q`
+  <!DOCTYPE cds [
+  <!ELEMENT cds (album+)>
+  <!ATTLIST cds
+            creator  CDATA
+            date     CDATA #REQUIRED
+            type     (a|b|c) #REQUIRED "a"
+  >
+  <!ELEMENT album (#PCDATA)>
+  ]>
+  ` ,
+  no_delete => 1 ,
+  );
+
+
 =head2  args()
 
 Return the arguments names (not nodes).
@@ -1441,6 +1583,10 @@ B<Options:>
 
 =over 10
 
+=item nodtd
+
+Do not add in the XML content the DTD applied by the method I<apply_dtd()>.
+
 =item noident
 
 If set to true the data isn't idented.
@@ -1727,7 +1873,7 @@ not directly to the HASH tree inside the object, (This will fix wrong accesses):
   my $server = $XML->{server}[0] ; ## return $XML->{server}[0]
   my $server = $XML->{server}[1] ; ## return $XML->{server}[1]
 
-To get all the values of a multiple attribute/key:
+To get all the values of multiple elements/keys:
 
   ## This work having only a string inside {address}, or with an ARRAY ref:
   my @addrsses = @{$XML->{server}{address}} ;
@@ -1836,7 +1982,7 @@ Where TYPE can be:
 
 Example:
 
-  ## All the servers
+  ## A servers content
   my $name = $XML->{server}{name}('$') ;
   ## ... or:
   my $name = $XML->{server}{name}->content ;
@@ -1855,7 +2001,7 @@ Example:
   ## Without the reference:
   my @servers = $XML->{server}('$@') ;
   
-  ## Without the reference:
+  ## A XML data, same as data_pointer():
   my $xml_data = $XML->{server}('<xml>') ;
 
 
@@ -1939,8 +2085,9 @@ detected using this roles:
     
     \s \w \d
     !"#$%&'()*+,-./:;<=>?@[\]^`{|}~
-    
-    
+    0x80, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8e, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9e, 0x9f, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff, 0x20, 
+
+  TODO: 0x81, 0x8d, 0x8f, 0x90, 0xa0 
   
   CDATA:
   - If have tags: <...>
@@ -2197,12 +2344,54 @@ L<XML.com|http://www.xml.com>
 
 =head1 AUTHOR
 
-Graciliano M. P. <gm@virtuasites.com.br>
+Graciliano M. P. C<< <gm at virtuasites.com.br> >>
 
 I will appreciate any type of feedback (include your opinions and/or suggestions). ;-P
 
 Enjoy and thanks for who are enjoying this tool and have sent e-mails! ;-P
 
+=head1 CURRENT MAINTAINER
+
+Harish Madabushi, C<< <harish.tmh at gmail.com> >>
+
+=head1 BUGS
+
+Please report any bugs or feature requests to C<bug-xml-smart at rt.cpan.org>, or through
+the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=XML-Smart>.  Both the author and the
+maintainer will be notified, and then you'll automatically be notified of progress on your bug as changes are made.
+
+=head1 SUPPORT
+
+You can find documentation for this module with the perldoc command.
+
+    perldoc XML::Smart
+
+You can also look for information at:
+
+=over 5
+
+=item * RT: CPAN's request tracker (report bugs here)
+
+L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=XML-Smart>
+
+=item * AnnoCPAN: Annotated CPAN documentation
+
+L<http://annocpan.org/dist/XML-Smart>
+
+=item * CPAN Ratings
+
+L<http://cpanratings.perl.org/d/XML-Smart>
+
+=item * Search CPAN
+
+L<http://search.cpan.org/dist/XML-Smart/>
+
+=item * GitHub CPAN
+
+L<https://github.com/harishmadabushi/XML-Smart>
+
+=back
+
 =head1 THANKS
 
 Thanks to Rusty Allen for the extensive tests of CDATA and BINARY handling of XML::Smart.
@@ -2211,7 +2400,9 @@ Thanks to Ted Haining to point a Perl-5.8.0 bug for tied keys of a HASH.
 
 Thanks to everybody that have sent ideas, patches or pointed bugs.
 
-=head1 COPYRIGHT
+=head1 LICENSE AND COPYRIGHT
+
+Copyright 2003 Graciliano M. P.
 
 This program is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
