@@ -2,7 +2,7 @@
 ## Name:        Data.pm
 ## Purpose:     XML::Smart::Data - Generate XML data.
 ## Author:      Graciliano M. P.
-## Modified by:
+## Modified by: Harish Madabushi
 ## Created:     28/09/2003
 ## RCS-ID:      
 ## Copyright:   (c) 2003 Graciliano M. P.
@@ -10,27 +10,32 @@
 ##              modify it under the same terms as Perl itself
 #############################################################################
 
-package XML::Smart::Data ;
+package XML::Smart::Data                                       ;
+
+
+use strict                                                     ;
+use warnings                                                   ;
+
+require Exporter                                               ;
+
+use XML::Smart::Entity qw(_add_basic_entity)                   ;
+use XML::Smart::Shared qw( _unset_sig_warn _reset_sig_warn )   ;
+
 
 our ($VERSION , @ISA) ;
-$VERSION = '0.01' ;
+$VERSION = '0.03' ;
 
-require Exporter ;
 @ISA = qw(Exporter) ;
 
 our @EXPORT = qw(data) ;
 our @EXPORT_OK = @EXPORT ;
 
-use strict ;
-no warnings ;
-
-use XML::Smart::Entity qw(_add_basic_entity) ;
-
 ########
 # DATA #
 ########
 
 sub data {
+  _unset_sig_warn() ;
   my $this = shift ;
   my ( %args ) = @_ ;
   
@@ -42,7 +47,7 @@ sub data {
   {
     my $addroot ;
 
-    if ( ref $tree ne 'HASH' ) { $addroot = 1 ;}
+    if ( $args{root} || ref $tree ne 'HASH' ) { $addroot = 1 ;}
     else {
       my $ks = keys %$tree ;
       my $n = 1 ;
@@ -87,7 +92,7 @@ sub data {
   }
 
   my $enc = 'iso-8859-1' ;
-  if ($unicode) { $enc = 'utf-8' ;}
+  if ($unicode) { $enc = 'UTF-8' ;}
     
   my $meta ;
   if ( $args{meta} ) {
@@ -130,11 +135,23 @@ sub data {
   
   if ( $args{noheader} ) { $xml = '' ; $metagen = '' if $args{nometagen} eq '' ;}
   
-  $data = $xml . $metagen . $meta . $data ;
+  my $dtd ;
+  
+  if ( !$args{nodtd} && $$this->{DTD} ) {
+    $dtd = ref $$this->{DTD} ? $$this->{DTD}->CutDTD : $$this->{DTD} ;
+    $dtd =~ s/\s*$// ;
+    $dtd = "\n$dtd" if $dtd ne '' && !$args{nospace} ;
+  }
+  
+  $data = $xml . $metagen . $meta . $dtd . $data ;
   
   if ($xml eq '') { $data =~ s/^\s+//gs ;}
   
-  if (wantarray) { return($data , $unicode) ;}
+  if (wantarray) { 
+      _reset_sig_warn() ;
+      return($data , $unicode) ;
+  }
+  _reset_sig_warn() ;
   return($data) ;
 }
 
@@ -143,6 +160,7 @@ sub data {
 #################
 
 sub is_valid_tree {
+  _unset_sig_warn() ;
   my ( $tree ) = @_ ;
   my $found ;
   if (ref($tree) eq 'HASH') {
@@ -162,6 +180,7 @@ sub is_valid_tree {
   }
   elsif (ref($tree) eq 'SCALAR' && $$tree ne '') { $found = 1 ;}
   
+  _reset_sig_warn() ;
   return $found ;
 }
 
@@ -170,17 +189,37 @@ sub is_valid_tree {
 ###############
 
 sub _is_unicode {
-  if ($] >= 5.8) {
-    eval(q`
-      if ( $data =~ /[\x{100}-\x{10FFFF}]/s) { return 1 ;}}
+  _unset_sig_warn() ;
+  if ($] >= 5.008001) {
+    if ( utf8::is_utf8($_[0])) { 
+	_reset_sig_warn() ;
+	return 1 ;
+    }
+  }
+  elsif ($] >= 5.008) {
+    require Encode ;
+    if ( Encode::is_utf8($_[0])) { 
+	_reset_sig_warn() ;
+	return 1 ;
+    }
+  }
+  elsif ( $] >= 5.007 ) {
+    _reset_sig_warn() ;
+    my $is = eval(q`
+      if ( $_[0] =~ /[\x{100}-\x{10FFFF}]/s) { return 1 ;}
+      return undef ;
     `);
+    $@ = undef ;
+    return 1 if $is ;
   }
   else {
     ## No Perl internal support for UTF-8! ;-/
     ## Is better to handle as Latin1.
+    _reset_sig_warn() ;
     return undef ;
   }
 
+  _reset_sig_warn() ;
   return undef ;
 }
 
@@ -189,12 +228,19 @@ sub _is_unicode {
 #########
 
 sub _data {
+
+  _unset_sig_warn() ;
   my ( $data , $tree , $tag , $level , $prev_tree , $parsed , $ar_i , $node_type , @stat ) = @_ ;
 
-  if (ref($tree) eq 'XML::Smart') { $tree = $$tree->{point} ;}
+  if (ref($tree) eq 'XML::Smart') { $tree = defined $$tree->{content} ? $$tree->{content} : $$tree->{point} ;}
   
-  if ($$parsed{"$tree"}) { return ;}
-  $$parsed{"$tree"}++ ;
+  if ( ref($tree) ) {
+    if ($$parsed{"$tree"}) { 
+	_reset_sig_warn() ;
+	return ;
+    }
+    ++$$parsed{"$tree"} ;
+  }
   
   my $ident = "\n" ;
   $ident .= '  ' x $level if !$stat[0] ;
@@ -215,7 +261,7 @@ sub _data {
     if ( !$stat[5] && $tree->{'/order'} ) {
       my %keys ;
       foreach my $keys_i ( @{$tree->{'/order'}} ) {
-        if ( exists $$tree{$keys_i} && (!ref($$tree{$keys_i}) || ref($$tree{$keys_i}) eq 'HASH' || (ref($$tree{$keys_i}) eq 'ARRAY' && exists $$tree{$keys_i}[ $keys{$keys_i} ] ) ) ) {        
+        if ( exists $$tree{$keys_i} && (!ref($$tree{$keys_i}) || ref($$tree{$keys_i}) eq 'HASH' || ref($$tree{$keys_i}) eq 'XML::Smart' || (ref($$tree{$keys_i}) eq 'ARRAY' && exists $$tree{$keys_i}[ $keys{$keys_i} ] ) ) ) {
           push(@all_keys , $keys_i) ;
           
           if ( ++$keys{$keys_i} == 2 && ref $$tree{$keys_i} eq 'ARRAY' ) {
@@ -234,7 +280,7 @@ sub _data {
 
     foreach my $Key ( @all_keys ) {
       if ($Key eq '' || $Key eq '/order' || $Key eq '/nodes') { next ;}
-            
+
       if ( $Key eq '!--' && (!ref($$tree{$Key}) || ( ref($$tree{$Key}) eq 'HASH' && (keys %{$$tree{$Key}}) == 1 && (defined $$tree{$Key}{CONTENT} || defined $$tree{$Key}{content}) ) ) ) {
         my $ct = $$tree{$Key} ;
         if (ref $$tree{$Key}) { $ct = defined $$tree{$Key}{CONTENT} ? $$tree{$Key}{CONTENT} : $$tree{$Key}{content} ;} ;
@@ -243,7 +289,9 @@ sub _data {
       elsif (ref($$tree{$Key})) {
         my $k = $$tree{$Key} ;
         my $i ;
-        if (ref $k eq 'XML::Smart') { $k = ${$$tree{$Key}}->{point} ;}
+        if (ref $k eq 'XML::Smart') {
+          $k = defined ${$$tree{$Key}}->{content} ? ${$$tree{$Key}}->{content} : ${$$tree{$Key}}->{point} ;
+        }
         elsif ( ref $k eq 'ARRAY' && $multi_keys{$Key} ) {
           $i = $array_i{$Key}++ if $#{$k} > 0 ;
         }
@@ -254,7 +302,7 @@ sub _data {
         my $k = [$$tree{$Key}] ;
         $args .= &_data(\$tags,$k,$Key, $level+1 , $tree , $parsed , undef , $$tree{'/nodes'}{$Key} , @stat) ;
       }
-      elsif ("\L$Key\E" eq 'content') {
+      elsif (lc($Key) eq 'content') {
         if ( tied($$tree{$Key}) && $$tree{$Key} =~ /\S/s ) {
           $ident = '' ; $stat[1] += 2 ;
         }
@@ -263,7 +311,7 @@ sub _data {
         if ( $$tree{$Key} ne '' ) {
           my $p0 = length($tags) ;
           $tags .= $$tree{$Key} ;        
-          $cont = [$p0,length($tags)] ;
+          $cont = [$p0, length($tags) - $p0] ;
         }
       }
       elsif ($Key =~ /^\/\.CONTENT\/\d+$/) { $tags .= $$tree{$Key} ;}
@@ -299,13 +347,12 @@ sub _data {
     foreach my $Key ( keys %array_i ) {
       if ( $array_i{$Key} ne 'ok' && $#{ $$tree{$Key} } >= $array_i{$Key} ) {
         for my $i ( $array_i{$Key} .. $#{ $$tree{$Key} } ) {
-          my $k = $$tree{$Key}[$i] ;
-          $args .= &_data(\$tags,$k,$Key, $level+1 , $tree , $parsed , undef , undef , @stat) ;
+          $args .= &_data(\$tags, $$tree{$Key} ,$Key, $level+1 , $tree , $parsed , $i , $$tree{'/nodes'}{$Key} , @stat) ;
         }
       }
     }
     
-    if ( $cont ) {
+    if ( $cont ne '' ) {
       my ( $po , $p1 ) = @$cont ;
       my $cont = substr($tags , $po , $p1) ;
         
@@ -333,14 +380,17 @@ sub _data {
       }
       else { &_add_basic_entity($cont) ;}
       
-      my $px = $p1 ;
+      my $pe = $po + $p1 ;
+      my $px = $pe ;
       while( substr($tags , $px , 1) =~ /\s/ ) { ++$px ;}
-      
-      if ( $px > $p1 ) { substr($tags , $p1 , $px-$p1) = '' ;}
+
+      if ( $px > $pe ) { substr($tags , $pe , $px-$pe) = '' ;}
       
       substr($tags , $po , $p1) = $cont ;
     }
     
+    ##print "***$tag>> $args,$args_end,$tags,$cont,$stat_1 [@all_keys]\n" ;
+    
     if ($args_end ne '') {
       $args .= $args_end ;
       $args_end = undef ;
@@ -364,17 +414,14 @@ sub _data {
       $$data .= $ident if !$cont ;
       $$data .= qq`</$tag>` if $tag ne '' ;
     }
+    else {
+      $$data .= qq`$ident<$tag></$tag>` if $tag ne '' ;
+    }
   }
   elsif (ref($tree) eq 'ARRAY') {
     my ($c,$v,$tags) ;
 
-    my $i = -1 ;
-    foreach my $value_i (@$tree) {
-      
-      if ( $ar_i ne '' ) {
-        ++$i ;
-        next if $i != $ar_i ;
-      }
+    foreach my $value_i ( ($ar_i ne '' ? $$tree[$ar_i] : @$tree) ) {
       
       my $value = $value_i ;
       if (ref $value_i eq 'XML::Smart') { $value = $$value_i->{point} ;}
@@ -439,8 +486,10 @@ sub _data {
       my $k = $stat[4] ? $tag : &_check_key($tag) ;
       if    ($stat[3] == 1) { $k = "\L$k\E" ;}
       elsif ($stat[3] == 2) { $k = "\U$k\E" ;}
-      delete $$parsed{"$tree"} ;
-      return " $k=" . &_add_quote($v) ;
+      delete $$parsed{"$tree"} if ref($tree) ;
+      my $return_val = " $k=" . &_add_quote($v)  ;
+      _reset_sig_warn() ;
+      return $return_val ;
     }
     else { $$data .= $tags ;}
   }
@@ -448,18 +497,32 @@ sub _data {
     my $k = $stat[4] ? $tag : &_check_key($tag) ;
     if    ($stat[3] == 1) { $k = "\L$k\E" ;}
     elsif ($stat[3] == 2) { $k = "\U$k\E" ;}
-    delete $$parsed{"$tree"} ;
-    return " $k=" . &_add_quote($$tree) ;
+    delete $$parsed{"$tree"} if ref($tree) ;
+    my $return_val = " $k=" . &_add_quote($$tree) ;
+    _reset_sig_warn() ;
+    return $return_val ;
   }
   elsif (ref($tree)) {
     my $k = $stat[4] ? $tag : &_check_key($tag) ;
     if    ($stat[3] == 1) { $k = "\L$k\E" ;}
     elsif ($stat[3] == 2) { $k = "\U$k\E" ;}
-    delete $$parsed{"$tree"} ;
-    return " $k=" . &_add_quote("$tree") ;
+    delete $$parsed{"$tree"} if ref($tree) ;
+    my $return_val = " $k=" . &_add_quote("$tree") ;
+    _reset_sig_warn() ;
+    return $return_val ;
+  }
+  else {
+    my $k = $stat[4] ? $tag : &_check_key($tag) ;
+    if    ($stat[3] == 1) { $k = "\L$k\E" ;}
+    elsif ($stat[3] == 2) { $k = "\U$k\E" ;}
+    delete $$parsed{"$tree"} if ref($tree) ;
+    my $return_val = " $k=" . &_add_quote($tree) ;
+    _reset_sig_warn() ;
+    return $return_val ;
   }
 
-  delete $$parsed{"$tree"} ;
+  delete $$parsed{"$tree"} if ref($tree) ;
+  _reset_sig_warn() ;
   return ;
 }
 
@@ -485,13 +548,16 @@ sub _check_tag { &_check_key ;}
 ##############
 
 sub _check_key {
+  _unset_sig_warn() ;
   if ($_[0] =~ /(?:^[.:-]|[^\w\:\.\-])/s) {
     my $k = $_[0] ;
     $k =~ s/^[.:-]+//s ;
     $k =~ s/[^\w\:\.\-]+/_/gs ;
     return( $k ) ;
   }
-  return( $_[0] ) ;
+  my $return_val = $_[0] ;
+  _reset_sig_warn() ;
+  return( $return_val ) ;
 }
 
 ##############
@@ -499,6 +565,7 @@ sub _check_key {
 ##############
 
 sub _add_quote {
+  _unset_sig_warn() ;
   my ($data) = @_ ;
   $data =~ s/\\$/\\\\/s ;
   
@@ -507,17 +574,29 @@ sub _add_quote {
   my $q1 = ($data =~ /"/s) ? 1 : undef ;
   my $q2 = ($data =~ /'/s) ? 1 : undef ;
   
-  if (!$q1 && !$q2) { return( qq`"$data"` ) ;}
+  if (!$q1 && !$q2) { 
+      _reset_sig_warn() ;
+      return( qq`"$data"` ) ;
+  }
   
   if ($q1 && $q2) {
     $data =~ s/"/&quot;/gs ;
+    _reset_sig_warn() ;
     return( qq`"$data"` ) ;
   }
   
-  if ($q1) { return( qq`'$data'` ) ;}
-  if ($q2) { return( qq`"$data"` ) ;}
+  if ($q1) { 
+      _reset_sig_warn() ;
+      return( qq`'$data'` ) ;
+  }
+  if ($q2) { 
+      _reset_sig_warn() ;
+      return( qq`"$data"` ) ;
+  }
 
+  _reset_sig_warn() ;
   return( qq`"$data"` ) ;
+
 }
 
 #######
