@@ -2,7 +2,7 @@
 ## Name:        Tree.pm
 ## Purpose:     XML::Smart::Tree
 ## Author:      Graciliano M. P.
-## Modified by:
+## Modified by: Harish Madabushi
 ## Created:     10/05/2003
 ## RCS-ID:      
 ## Copyright:   (c) 2003 Graciliano M. P.
@@ -10,57 +10,56 @@
 ##              modify it under the same terms as Perl itself
 #############################################################################
 
-package XML::Smart::Tree ;
+package XML::Smart::Tree                                       ;
 
-use XML::Smart::Data ;
+use strict                                                     ;
+use warnings                                                   ;
 
-no warnings ;
+use XML::Smart::Entity qw(_parse_basic_entity)                 ;
+use XML::Smart::Shared qw( _unset_sig_warn _reset_sig_warn )   ;
 
 our ($VERSION) ;
-$VERSION = '1.0' ;
+$VERSION = '1.1' ;
+
+my %PARSERS = (
+    XML_Parser => 0 ,
+    XML_Smart_Parser => 0 ,
+    XML_Smart_HTMLParser => 0 ,
+    ) ;
+
+my $DEFAULT_LOADED ;
+
+use vars qw($NO_XML_PARSER);
 
-  my %PARSERS = (
-  XML_Parser => 0 ,
-  XML_Smart_Parser => 0 ,
-  XML_Smart_HTMLParser => 0 ,
-  ) ;
-  
-  my $DEFAULT_LOADED ;
-  
-  use vars qw($NO_XML_PARSER);
 
 ###################
 # LOAD_XML_PARSER #
 ###################
 
 sub load_XML_Parser {
-  return if $NO_XML_PARSER ;
-  
-  eval('use XML::Parser ;') ;
-  if ($@) { return( undef ) ;}
-  
-  my ($xml , $tree) ;
-  
-  eval {
-    my $sig_warn = $SIG{__WARN__} ;
-    my $sig_die = $SIG{__DIE__} ;
+
+    return if $NO_XML_PARSER ;
     
-    $SIG{__WARN__} = sub {} ;
-    $SIG{__DIE__} = sub {} ;
+    _unset_sig_warn() ;
+    eval('use XML::Parser ;') ;
+    _reset_sig_warn() ; 
+    if ($@) { $@ = undef ; return( undef ) ;}
     
-    no strict ;
-  
-    my $data = '<root><foo arg1="t1" arg2="t2" /></root>' ;
-    $xml = XML::Parser->new(Style => 'Tree') ;
-    $tree = $xml->parse($data) ;
+    my ($xml , $tree) ;
     
-    $SIG{__WARN__} = $sig_warn ;
-    $SIG{__DIE__} = $sig_die ;
-  } ;
+    _unset_sig_warn() ;
+    eval {
+	no strict ;
+	my $data = '<root><foo arg1="t1" arg2="t2" /></root>' ;
+	$xml = XML::Parser->new(Style => 'Tree') ;
+	$tree = $xml->parse($data) ;
+    } ;
+    _reset_sig_warn() ;
   
-  if (!$tree || ref($tree) ne 'ARRAY') { return( undef ) ;}
-  if ($tree->[1][2][0]{arg1} eq 't1') { return( 1 ) ;}
-  return( undef ) ;
+    if (!$tree || ref($tree) ne 'ARRAY') { return( undef ) ;}
+    if ($tree->[1][2][0]{arg1} eq 't1') { return( 1 ) ;}
+    return( undef ) ;
+
 }
 
 #########################
@@ -68,9 +67,13 @@ sub load_XML_Parser {
 #########################
 
 sub load_XML_Smart_Parser {
+
+  _unset_sig_warn() ;
   eval('use XML::Smart::Parser ;') ;
-  if ($@) { return( undef ) ;}
+  _reset_sig_warn() ;
+  if ($@) { $@ = undef ; return( undef ) ;}
   return(1) ;
+
 }
 
 #############################
@@ -78,8 +81,10 @@ sub load_XML_Smart_Parser {
 #############################
 
 sub load_XML_Smart_HTMLParser {
-  eval('use XML::Smart::HTMLParser ;') ;
-  if ($@) { return( undef ) ;}
+  _unset_sig_warn() ;
+    eval('use XML::Smart::HTMLParser ;') ;
+  _reset_sig_warn() ;
+  if ($@) { $@ = undef ; return( undef ) ;}
   return(1) ;
 }
 
@@ -95,7 +100,7 @@ sub load {
     $parser =~ s/:+/_/gs ;
     $parser =~ s/\W//g ;
     
-    if    ($parser =~ /^html?$/i) { $parser = 'XML_Smart_HTMLParser' ;}
+    if    ($parser =~ /^(?:html?|wild)$/i) { $parser = 'XML_Smart_HTMLParser' ;}
     elsif ($parser =~ /^(?:re|smart)/i) { $parser = 'XML_Smart_Parser' ;}
     
     foreach my $Key ( keys %PARSERS ) {
@@ -104,27 +109,25 @@ sub load {
   }
   
   my $ok ;
-  if ($module eq 'XML_Parser') {
+  if( $module && ( $module eq 'XML_Parser' ) ) {
     $PARSERS{XML_Parser} = 1 if &load_XML_Parser() ;
     $ok = $PARSERS{XML_Parser} ;
-  }
-  elsif ($module eq 'XML_Smart_Parser') {
-    $PARSERS{XML_Smart_Parser} = 1 if !$PARSERS{XML_Smart_Parser} && &load_XML_Smart_Parser() ;
-    $ok = $PARSERS{XML_Smart_Parser} ;
-  }
-  elsif ($module eq 'XML_Smart_HTMLParser') {
-    $PARSERS{XML_Smart_HTMLParser} = 1 if !$PARSERS{XML_Smart_HTMLParser} && &load_XML_Smart_HTMLParser() ;
-    $ok = $PARSERS{XML_Smart_HTMLParser} ;
+  } elsif ( $module && ( $module eq 'XML_Smart_Parser' ) ) {
+      $PARSERS{XML_Smart_Parser} = 1 if !$PARSERS{XML_Smart_Parser} && &load_XML_Smart_Parser() ;
+      $ok = $PARSERS{XML_Smart_Parser} ;
+  } elsif( $module and ( $module eq 'XML_Smart_HTMLParser' ) ) {
+      $PARSERS{XML_Smart_HTMLParser} = 1 if !$PARSERS{XML_Smart_HTMLParser} && &load_XML_Smart_HTMLParser() ;
+      $ok = $PARSERS{XML_Smart_HTMLParser} ;
   }
   
   if (!$ok && !$DEFAULT_LOADED) {
-    $PARSERS{XML_Parser} = 1 if &load_XML_Parser() ;
-    $module = 'XML_Parser' ;
-    if ( !$PARSERS{XML_Parser} ) {
-      $PARSERS{XML_Smart_Parser} = 1 if &load_XML_Smart_Parser() ;  
-      $module = 'XML_Smart_Parser' ;
-    }
-    $DEFAULT_LOADED = 1 ;
+      $PARSERS{XML_Parser} = 1 if &load_XML_Parser() ;
+      $module = 'XML_Parser' ;
+      if ( !$PARSERS{XML_Parser} ) {
+	  $PARSERS{XML_Smart_Parser} = 1 if &load_XML_Smart_Parser() ;  
+	  $module = 'XML_Smart_Parser' ;
+      }
+      $DEFAULT_LOADED = 1 ;
   }
   
   return($module) ;
@@ -166,14 +169,22 @@ sub parse {
   else { croak("Can't find a parser for XML!") ;}
   
   shift(@_) ;
-  if ( $_[0] =~ /^\s*(?:XML_\w+|html?|re\w+|smart)\s*$/i) { shift(@_) ;}
+  if ( $_[0] && ( $_[0] =~ /^\s*(?:XML_\w+|html?|re\w+|smart)\s*$/i ) ) { shift(@_) ;}
 
+  _unset_sig_warn() ;
   my ( %args ) = @_ ;
+  _reset_sig_warn() ;
   
   if ( $args{lowtag} ) { $xml->{SMART}{tag} = 1 ;}
   if ( $args{upertag} ) { $xml->{SMART}{tag} = 2 ;}
   if ( $args{lowarg} ) { $xml->{SMART}{arg} = 1 ;}
   if ( $args{uperarg} ) { $xml->{SMART}{arg} = 2 ;}
+  if ( $args{arg_single} ) { $xml->{SMART}{arg_single} = 1 ;}  
+
+  if ( $args{no_order} ) { $xml->{SMART}{no_order} = 1 ;}
+  if ( $args{no_nodes} ) { $xml->{SMART}{no_nodes} = 1 ;}
+  
+  if ( $args{use_spaces} ) { $xml->{SMART}{use_spaces} = 1 ;}
   
   $xml->{SMART}{on_start} = $args{on_start} if ref($args{on_start}) eq 'CODE' ;
   $xml->{SMART}{on_char}  = $args{on_char}  if ref($args{on_char})  eq 'CODE' ;
@@ -247,15 +258,26 @@ sub _Init {
 ##########
 
 sub _Start {
-  my $this = shift ; ##print "START>> @_\n" ;
-  my ($tag , %args) = @_ ;
+  my $this = shift ;
   
-  if    ( $this->{SMART}{tag} == 1 ) { $tag = lc($tag) ;}
-  elsif ( $this->{SMART}{tag} == 2 ) { $tag = uc($tag) ;}
+  if ( $this->{LAST_CALL} && ( $this->{LAST_CALL} eq 'char' ) ) { 
+      _Char_process( $this , delete $this->{CONTENT_BUFFER} ) ;
+  }
   
-  $this->{PARSING}{p}{'/nodes'}{$tag} = 1 ;
+  ##print "START>> @_\n" ;
   
-  push( @{$this->{PARSING}{p}{'/order'}} , $tag) ;
+  $this->{LAST_CALL} = 'start' ;
+
+  _unset_sig_warn();
+  my ( $tag , %args ) = @_ ;
+  _reset_sig_warn();
+
+  if    ( $this->{SMART}{tag} && ( $this->{SMART}{tag} == 1 ) ) { $tag = lc($tag) ;}
+  elsif ( $this->{SMART}{tag} && ( $this->{SMART}{tag} == 2 ) ) { $tag = uc($tag) ;}
+  
+  $this->{PARSING}{p}{'/nodes'}{$tag} = 1 if !$this->{SMART}{no_nodes} ;
+  
+  push( @{$this->{PARSING}{p}{'/order'}} , $tag) if !$this->{SMART}{no_order} ;
   
   if ( $this->{SMART}{arg} ) {
     my $type = $this->{SMART}{arg} ;
@@ -278,8 +300,14 @@ sub _Start {
     %args = %argsok ;
   }
   
+  if ( $this->{SMART}{arg_single} ) {
+    foreach my $Key ( keys %args ) {
+      $args{$Key} = 1 if !defined $args{$Key} ;
+    }
+  }
+  
   ## Args order:
-  {
+  if ( !$this->{SMART}{no_order} ) {
     my @order ; 
     for(my $i = 1 ; $i < $#_ ; $i+=2) { push( @order , $_[$i] ) ;}
     
@@ -298,7 +326,7 @@ sub _Start {
   $args{'/back'} = $this->{PARSING}{p} ;
   
   if ($this->{NOENTITY}) {
-    foreach my $Key ( keys %args ) { &XML::Smart::Data::_parse_basic_entity( $args{$Key} ) ;}
+    foreach my $Key ( keys %args ) { &_parse_basic_entity( $args{$Key} ) ;}
   }
   
   if ( defined $this->{PARSING}{p}{$tag} ) {
@@ -321,7 +349,7 @@ sub _Start {
   
   if ( $this->{SMART}{on_start} ) {
     my $sub = $this->{SMART}{on_start} ;
-    &$sub($tag , $this->{PARSING}{p} , $this->{PARSING}{p}{'/back'}) ;
+    &$sub($tag , $this->{PARSING}{p} , $this->{PARSING}{p}{'/back'} , undef , $this ) ;
   }
   
   return ;
@@ -330,27 +358,43 @@ sub _Start {
 #########
 # _CHAR #
 #########
+#
+# XML::Parser parse each line as a different call to _Char().
+# For XML::Smart multiple calls to _Char() occurs only when the content
+# have other nodes inside.
+#
 
-sub _Char {
+sub _Char { ##print "CHAR>>\n" ;
   my $this = shift ;
-  #print "CONT>> ##@_##\n" ;
+  $this->{CONTENT_BUFFER} .= $_[0] ;
+  $this->{LAST_CALL} = 'char' ;
+  return ;
+}
+
+sub _Char_process {
+  my $this = shift ;
+  ##print "CONT>> ##@_##\n" ;
 
   my $content = $_[0] ;
   
+  if ( !$this->{SMART}{use_spaces} && $content !~ /\S+/s ) { return ;}
+
+  ######
+  
   if (! defined $this->{PARSING}{p}{'dt:dt'} && defined $this->{PARSING}{p}{'DT:DT'}) {
     $this->{PARSING}{p}{'dt:dt'} = delete $this->{PARSING}{p}{'DT:DT'} ;
   }
   
-  if ( $content !~ /\S+/s ) { return ;}
-  
-  if ( $this->{PARSING}{p}{'dt:dt'} =~ /binary\.base64/si ) {
+  if ( $this->{PARSING}{p}{'dt:dt'} && ( $this->{PARSING}{p}{'dt:dt'} =~ /binary\.base64/si ) ) {
     require XML::Smart::Base64 ;
     $content = &XML::Smart::Base64::decode_base64($content) ;
     delete $this->{PARSING}{p}{'dt:dt'} ;
-    delete $this->{PARSING}{p}{'/nodes'}{'dt:dt'} ;
-
-    my $nkeys = keys %{$this->{PARSING}{p}{'/nodes'}} ;
-    if ($nkeys < 1) { delete $this->{PARSING}{p}{'/nodes'} ;}
+    
+    if ( $this->{PARSING}{p}{'/nodes'} ) {
+      delete $this->{PARSING}{p}{'/nodes'}{'dt:dt'} ;
+      my $nkeys = keys %{$this->{PARSING}{p}{'/nodes'}} ;
+      if ($nkeys < 1) { delete $this->{PARSING}{p}{'/nodes'} ;}
+    }
     
     if ( $this->{PARSING}{p}{'/order'} ) {
       my @order = @{$this->{PARSING}{p}{'/order'}} ;
@@ -360,11 +404,13 @@ sub _Char {
       else { delete $this->{PARSING}{p}{'/order'} ;}
     }
   }
-  elsif ($this->{NOENTITY}) { &XML::Smart::Data::_parse_basic_entity($content) ;}
+  elsif ($this->{NOENTITY}) { &_parse_basic_entity($content) ;}
+  
+  ######
   
   if ( !exists $this->{PARSING}{p}{CONTENT} ) {
     $this->{PARSING}{p}{CONTENT} = $content ;
-    push(@{$this->{PARSING}{p}{'/order'}} , 'CONTENT') ;
+    push(@{$this->{PARSING}{p}{'/order'}} , 'CONTENT') if !$this->{SMART}{no_order} ;
   }
   else {
     if ( !tied $this->{PARSING}{p}{CONTENT} ) {
@@ -376,19 +422,23 @@ sub _Char {
       $this->{PARSING}{p}{'/.CONTENT/x'} = 0 ;
       $this->{PARSING}{p}{"/.CONTENT/0"} = $cont ;
       
-      splice( @{$this->{PARSING}{p}{'/order'}} , -1,0, "/.CONTENT/0") ; # Add inside (0=init line , 1=end line).
+      my $cont_pos = 0 ;
+      for my $key ( @{$this->{PARSING}{p}{'/order'}} ) {
+        last if ($key eq 'CONTENT') ;
+        ++$cont_pos ;
+      }
       
-      push( @{$this->{PARSING}{p}{'/order'}} , "/.CONTENT/0") ;
+      splice( @{$this->{PARSING}{p}{'/order'}} , $cont_pos,0, "/.CONTENT/0") if !$this->{SMART}{no_order} ;
     }
 
     my $x = ++$this->{PARSING}{p}{'/.CONTENT/x'} ;
     $this->{PARSING}{p}{"/.CONTENT/$x"} = $content ;
-    push( @{$this->{PARSING}{p}{'/order'}} , "/.CONTENT/$x") ;
+    push( @{$this->{PARSING}{p}{'/order'}} , "/.CONTENT/$x") if !$this->{SMART}{no_order} ;
   }
   
   if ( $this->{SMART}{on_char} ) {
     my $sub = $this->{SMART}{on_char} ;
-    &$sub($this->{PARSING}{p}{'/tag'} , $this->{PARSING}{p} , $this->{PARSING}{p}{'/back'} , \$this->{PARSING}{p}{CONTENT}) ;
+    &$sub($this->{PARSING}{p}{'/tag'} , $this->{PARSING}{p} , $this->{PARSING}{p}{'/back'} , \$this->{PARSING}{p}{CONTENT} , $this ) ;
   }
   
   return ;
@@ -400,10 +450,14 @@ sub _Char {
 
 sub _End { ##print "END>> @_[1] >> $_[0]->{PARSING}{p}{'/tag'}\n" ;
   my $this = shift ;
+  
+  if ( $this->{LAST_CALL} eq 'char' ) { _Char_process( $this , delete $this->{CONTENT_BUFFER} ) ;}
+  $this->{LAST_CALL} = 'end' ;
+  
   my $tag = shift ;
   
-  if    ( $this->{SMART}{tag} == 1 ) { $tag = lc($tag) ;}
-  elsif ( $this->{SMART}{tag} == 2 ) { $tag = uc($tag) ;}
+  if    ( $this->{SMART}{tag} && ( $this->{SMART}{tag} == 1 ) ) { $tag = lc($tag) ;}
+  elsif ( $this->{SMART}{tag} && ( $this->{SMART}{tag} == 2 ) ) { $tag = uc($tag) ;}
 
   if ( $this->{PARSING}{p}{'/tag'} ne $tag ) { return ;}
 
@@ -426,11 +480,11 @@ sub _End { ##print "END>> @_[1] >> $_[0]->{PARSING}{p}{'/tag'}\n" ;
   
   if ( $this->{SMART}{on_end} ) {
     my $sub = $this->{SMART}{on_end} ;
-    &$sub($tag , $this->{PARSING}{p} , $back) ;
+    &$sub($tag , $this->{PARSING}{p} , $back , undef , $this) ;
   }
 
   $this->{PARSING}{p} = $back ;
-  
+    
   return ;
 }
 
@@ -447,6 +501,7 @@ sub _Final {
   }
   
   delete $this->{TIED_CONTENTS} ;
+  delete $this->{LAST_CALL} ;
   
   delete($this->{PARSING}) ;
   return($tree) ;
@@ -458,3 +513,11 @@ sub _Final {
 
 1;
 
+
+__END__
+
+
+
+
+
+
