@@ -2,7 +2,7 @@
 ## Name:        XPath.pm
 ## Purpose:     XML::Smart::XPath - Compatibility with XPath (through XML::XPath).
 ## Author:      Graciliano M. P.
-## Modified by:
+## Modified by: Harish Madabushi
 ## Created:     01/10/2003
 ## RCS-ID:      
 ## Copyright:   (c) 2003 Graciliano M. P.
@@ -10,12 +10,20 @@
 ##              modify it under the same terms as Perl itself
 #############################################################################
 
-package XML::Smart::XPath ;
+package XML::Smart::XPath                                      ;
+
+use 5.006                                                      ;
+
+use strict                                                     ;
+use warnings                                                   ;
+
+require Exporter                                               ;
+
+use XML::Smart::Shared qw( _unset_sig_warn _reset_sig_warn )   ;
 
 our ($VERSION , @ISA) ;
-$VERSION = '0.01' ;
+$VERSION = '0.03' ;
 
-require Exporter ;
 @ISA = qw(Exporter) ;
 
 our @EXPORT = qw(xpath XPath xpath_pointer XPath_pointer) ;
@@ -59,6 +67,7 @@ sub xpath {
   }
   
   if ( !@_ ) { return $xpath ;}
+  return ;
 }
 
 #################
@@ -75,6 +84,7 @@ sub xpath_pointer {
   my $xpath = XML::XPath->new(xml => $this->data_pointer(nospace => 1 , noheader => 1)) ;
   
   if ( !@_ ) { return $xpath ;}
+  return ;
 }
 
 #######
